from abc import ABC, abstractmethod
from typing import Any
from psyki.logic.datalog.grammar import DatalogFormula, Expression, DefinitionClause, Argument, Negation, Unary, Nary, \
    Variable, Number, Predication
from psyki.resources.dist.DatalogParser import DatalogParser


class Adapter(ABC):

    @abstractmethod
    def get_formula(self, ast: Any) -> DatalogFormula:
        pass


class Antlr4(Adapter):

    def __init__(self):
        pass

    def get_formula(self, ast: DatalogParser.FormulaContext) -> DatalogFormula:
        return DatalogFormula(self._get_definition_clause(ast.lhs), self._get_clause(ast.rhs), '←')

    def _get_definition_clause(self, node: DatalogParser.DefPredicateArgsContext):
        return DefinitionClause(node.pred.text, self._get_arguments(node.args))

    def _get_arguments(self, node: DatalogParser.MoreArgsContext or DatalogParser.LastTermContext):
        if isinstance(node, DatalogParser.MoreArgsContext):
            return Argument(self._get_term(node.name), self._get_arguments(node.args))
        elif isinstance(node, DatalogParser.LastTermContext):
            return Argument(self._get_term(node.name))

    def _get_clause(self,
                    node: DatalogParser.ClauseExpressionContext or DatalogParser.ClauseExpressionNoParContext or DatalogParser.ClauseLiteralContext):
        if isinstance(node, DatalogParser.ClauseExpressionContext) \
                or isinstance(node, DatalogParser.ClauseExpressionNoParContext):
            return Expression(self._get_clause(node.left), self._get_clause(node.right), node.op.text)
        elif isinstance(node, DatalogParser.ClauseLiteralContext):
            return self._get_literal(node.lit)

    def _get_literal(self, node: DatalogParser.LiteralPredContext or DatalogParser.LiteralNegContext):
        if isinstance(node, DatalogParser.LiteralNegContext):
            return Negation(self._get_clause(node.pred))
        elif isinstance(node, DatalogParser.LiteralPredContext):
            return self._get_predicate(node.pred)

    def _get_predicate(self, node: DatalogParser.PredicateTermContext or DatalogParser.PredicateUnaryContext
                                   or DatalogParser.PredicateArgsContext):
        if isinstance(node, DatalogParser.PredicateTermContext):
            return self._get_term(node.name)
        elif isinstance(node, DatalogParser.PredicateUnaryContext):
            return Unary(node.pred.text)
        elif isinstance(node, DatalogParser.PredicateArgsContext):
            return Nary(node.pred.text, self._get_arguments(node.args))

    def _get_term(self, node: DatalogParser.TermVarContext or DatalogParser.TermConstContext):
        if isinstance(node, DatalogParser.TermVarContext):
            return Variable(node.var.text)
        elif isinstance(node, DatalogParser.TermConstContext):
            return self._get_constant(node.name)

    def _get_constant(self, node: DatalogParser.ConstNumberContext or DatalogParser.ConstNameContext):
        if isinstance(node, DatalogParser.ConstNumberContext):
            return Number(node.num.text)
        elif isinstance(node, DatalogParser.ConstNameContext):
            return Predication(node.name.text)



import csv
from pathlib import Path
from typing import TextIO
from test.resources import rules

PATH = Path(__file__).parents[0]


def get_rule_path(filename: str) -> Path:
    return PATH / f"{filename}.txt"


def open_rule(filename: str) -> TextIO:
    return open(get_rule_path(filename))


def get_rules(name: str) -> list[str]:
    result = []
    with open(str(rules.PATH / name) + '.txt', mode="r") as file:
        reader = csv.reader(file, delimiter=';')
        for item in reader:
            result += item
    return result



from __future__ import annotations
from abc import ABC
from typing import Any, Callable, List
from tensorflow import maximum, minimum, constant, SparseTensor, cast, reshape, reduce_max, tile, Tensor, gather
from tensorflow.keras.backend import to_dense
from tensorflow.keras.layers import Minimum, Maximum, Dense, Dot, Concatenate
from tensorflow.keras.layers import Lambda
from tensorflow.python.ops.array_ops import shape
from tensorflow.python.ops.init_ops import constant_initializer, Ones, Zeros
from psyki.logic.datalog.grammar import DatalogFormula, Expression, Variable, Number, Unary, Predication, \
    DefinitionClause, Argument, Nary, Negation
from psyki.ski import Fuzzifier, Formula
from psyki.utils import eta, eta_one_abs, eta_abs_one


class ConstrainingFuzzifier(Fuzzifier, ABC):
    pass


class StructuringFuzzifier(Fuzzifier, ABC):
    pass


class Lukasiewicz(ConstrainingFuzzifier):

    def __init__(self, class_mapping: dict[str, int], feature_mapping: dict[str, int]):
        super().__init__()
        self.feature_mapping = feature_mapping
        self.class_mapping = {string: cast(to_dense(SparseTensor([[0, index]], [1.], (1, len(class_mapping)))), float)
                              for string, index in class_mapping.items()}
        self.classes: dict[str, Callable] = {}
        self._predicates: dict[str, Callable] = {}
        self.__rhs: dict[str, Callable] = {}
        self.visit_mapping: dict[Formula.__class__, Callable] = {
            DatalogFormula: self._visit_formula,
            Expression: self._visit_expression,
            Negation: self._visit_negation,
            Variable: self._visit_variable,
            Number: self._visit_number,
            Unary: self._visit_unary,
            Nary: self._visit_nary
        }

    def visit(self, rules: List[Formula]) -> Any:
        for rule in rules:
            self._visit(rule)
        return self.classes

    def _visit(self, visitable: Formula) -> Any:
        return self.visit_mapping.get(visitable.__class__)(visitable)

    def _visit_formula(self, node: DatalogFormula):
        self._visit_definition_clause(node.lhs, self._visit(node.rhs))

    def _visit_definition_clause(self, node: DefinitionClause, r: Callable):
        definition_name = node.predication
        predication_name = self._get_predication_name(node.arg)

        if predication_name is not None:
            class_tensor = reshape(self.class_mapping[predication_name], (1, len(self.class_mapping)))
            l = lambda y: eta(reduce_max(abs(tile(class_tensor, (shape(y)[0], 1)) - y), axis=1))
            if predication_name not in self.classes.keys():
                self.classes[predication_name] = lambda x, y: eta(r(x) - l(y))
                self.__rhs[predication_name] = lambda x: r(x)
            else:
                incomplete_function = self.__rhs[predication_name]
                self.classes[predication_name] = lambda x, y: eta(minimum(incomplete_function(x), r(x)) - l(y))
                self.__rhs[predication_name] = lambda x: minimum(incomplete_function(x), r(x))
        else:
            if definition_name not in self._predicates.keys():
                self._predicates[definition_name] = lambda x: r(x)
            else:
                incomplete_function = self._predicates[definition_name]
                self._predicates[definition_name] = lambda x: eta(minimum(incomplete_function(x), r(x)))

    def _visit_expression(self, node: Expression):
        l, r = self._visit(node.lhs), self._visit(node.rhs)
        operation = {
            '∧': lambda x: eta(maximum(l(x), r(x))),
            '∨': lambda x: eta(minimum(l(x), r(x))),
            '→': lambda x: eta(l(x) - r(x)),
            '↔': lambda x: eta(abs(l(x) - r(x))),
            '=': lambda x: eta(abs(l(x) - r(x))),
            '<': lambda x: eta(constant(1.) - eta(minimum(eta(constant(1.) - maximum(l(x) - r(x))),
                                                          eta(abs(l(x) - r(x)))))),
            '≤': lambda x: eta(constant(1.) - eta(constant(1.) - maximum(constant(0.), l(x) - r(x)))),
            '>': lambda x: eta(constant(1.) - maximum(constant(0.), l(x) - r(x))),
            '≥': lambda x: eta(minimum(eta(constant(1.) - maximum(l(x) - r(x))), eta(abs(l(x) - r(x))))),
            'm': lambda x: minimum(l(x), r(x)),
            '+': lambda x: l(x) + r(x),
            '*': lambda x: l(x) * r(x)
        }
        return operation.get(node.op)

    def _visit_variable(self, node: Variable):
        return lambda x: x[:, self.feature_mapping[node.name]] if node.name in self.feature_mapping.keys() else None

    def _visit_number(self, node: Number):
        return lambda _: node.value

    def _visit_unary(self, node: Unary):
        return self._predicates[node.name]

    def _visit_nary(self, node: Nary):
        return self._predicates[node.name]

    def _visit_negation(self, node: Negation):
        return lambda x: eta(constant(1.) - self._visit(node.predicate)(x))

    def _get_predication_name(self, node: Argument):
        if node.arg is not None:
            return self._get_predication_name(node.arg)
        elif isinstance(node.term, Predication):
            return node.term.name
        else:
            return None


class SubNetworkBuilder(StructuringFuzzifier):

    def __init__(self, predictor_input: Tensor, feature_mapping: dict[str, int]):
        self.predictor_input = predictor_input
        self.feature_mapping = feature_mapping
        self.classes: dict[str, Tensor] = {}
        self._predicates: dict[str, Tensor] = {}
        self.__rhs: dict[str, Tensor] = {}
        self.visit_mapping: dict[Formula.__class__, Callable] = {
            DatalogFormula: self._visit_formula,
            Expression: self._visit_expression,
            Negation: self._visit_negation,
            Variable: self._visit_variable,
            Number: self._visit_number,
            Unary: self._visit_unary,
            Nary: self._visit_nary
        }

    def visit(self, rules: List[Formula]) -> Any:
        for rule in rules:
            self._visit(rule)
        return list(self.classes.values())

    def _visit(self, visitable: Formula) -> Any:
        return self.visit_mapping.get(visitable.__class__)(visitable)

    def _visit_formula(self, node: DatalogFormula):
        self._visit_definition_clause(node.lhs, self._visit(node.rhs))

    def _visit_definition_clause(self, node: DefinitionClause, r: Tensor):
        definition_name = node.predication
        predication_name = self._get_predication_name(node.arg)

        if predication_name is not None:
            if predication_name not in self.classes.keys():
                self.classes[predication_name] = r
                self.__rhs[predication_name] = r
            else:
                incomplete_function = self.__rhs[predication_name]
                self.classes[predication_name] = minimum(incomplete_function, r)
                self.__rhs[predication_name] = minimum(incomplete_function, r)
        else:
            if definition_name not in self._predicates.keys():
                self._predicates[definition_name] = r
            else:
                incomplete_function = self._predicates[definition_name]
                self._predicates[definition_name] = minimum(incomplete_function, r)

    def _visit_expression(self, node: Expression):
        previous_layer = [self._visit(node.lhs), self._visit(node.rhs)]
        operation = {
            '∧': Minimum()(previous_layer),
            '∨': Maximum()(previous_layer),
            '→': None,
            '↔': None,
            '=': Dense(1, kernel_initializer=constant_initializer([1, -1]),
                       activation=eta_one_abs, trainable=False)(Concatenate(axis=1)(previous_layer)),
            '<': Dense(1, kernel_initializer=constant_initializer([-1, 1]),
                       activation=eta, trainable=False)(Concatenate(axis=1)(previous_layer)),
            '≤': Maximum()([Dense(1, kernel_initializer=constant_initializer([-1, 1]),
                                  activation=eta, trainable=False)(Concatenate(axis=1)(previous_layer)),
                            Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                  activation=eta_one_abs, trainable=False)(Concatenate(axis=1)(previous_layer))]),
            '>': Dense(1, kernel_initializer=constant_initializer([1, -1]),
                       activation=eta, trainable=False)(Concatenate(axis=1)(previous_layer)),
            '≥': Maximum()([Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                  activation=eta, trainable=False)(Concatenate(axis=1)(previous_layer)),
                            Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                  activation=eta_one_abs, trainable=False)(Concatenate(axis=1)(previous_layer))]),
            'm': Minimum()(previous_layer),
            '+': Dense(1, kernel_initializer=Ones, activation='linear', trainable=False)(Concatenate(axis=1)
                                                                                         (previous_layer)),
            '*': Dot(axes=1)(previous_layer)
        }
        return operation.get(node.op)

    def _visit_variable(self, node: Variable):
        return Lambda(lambda x: gather(x, [self.feature_mapping[node.name]], axis=1))(self.predictor_input) if node.name in self.feature_mapping.keys() else None

    def _visit_number(self, node: Number):
        return Dense(1, kernel_initializer=Zeros,
                     bias_initializer=constant_initializer(node.value),
                     trainable=False, activation='linear')(self.predictor_input)

    def _visit_unary(self, node: Unary):
        return self._predicates[node.name]

    def _visit_nary(self, node: Nary):
        return self._predicates[node.name]

    def _visit_negation(self, node: Negation):
        return Dense(1, kernel_initializer=Ones, activation=eta_abs_one, trainable=False)(self._visit(node.predicate))

    def _get_predication_name(self, node: Argument):
        if node.arg is not None:
            return self._get_predication_name(node.arg)
        elif isinstance(node.term, Predication):
            return node.term.name
        else:
            return None



from pathlib import Path
from typing import Any
import numpy as np

PATH = Path(__file__).parents[0]


def get_dataset(name: str) -> Any:
    return np.genfromtxt(str(PATH / name) + '.csv', delimiter=',', dtype=float)



from typing import List
from numpy import argmax, array
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from tensorflow import Tensor
from tensorflow.keras.layers import Dense
from test.resources.data import get_dataset
from sklearn.metrics import f1_score

POKER_FEATURE_MAPPING = {
        'S1': 0,
        'R1': 1,
        'S2': 2,
        'R2': 3,
        'S3': 4,
        'R3': 5,
        'S4': 6,
        'R4': 7,
        'S5': 8,
        'R5': 9
    }

POKER_CLASS_MAPPING = {
        'nothing': 0,
        'pair': 1,
        'two': 2,
        'three': 3,
        'straight': 4,
        'flush': 5,
        'full': 6,
        'four': 7,
        'straight_flush': 8,
        'royal_flush': 9
    }


def get_mlp(input_layer: Tensor, output: int, layers: int, neurons: int, activation_function, last_activation_function):
    """
    Generate a NN with the given parameters
    """
    x = Dense(neurons, activation=activation_function)(input_layer)
    for i in range(2, layers):
        x = Dense(neurons, activation=activation_function)(x)
    return Dense(output, activation=last_activation_function)(x)


def get_processed_dataset(name: str, validation: float = 1.0):
    poker_training = get_dataset(name + '-training')
    poker_testing = get_dataset(name + '-testing')
    if validation < 1:
        _, poker_testing = train_test_split(poker_testing, test_size=validation, random_state=123,
                                            stratify=poker_testing[:, -1])
    train_x = poker_training[:, :-1]
    train_y = poker_training[:, -1]
    test_x = poker_testing[:, :-1]
    test_y = poker_testing[:, -1]

    # One Hot encode the class labels
    encoder = OneHotEncoder(sparse=False)
    encoder.fit_transform([train_y])
    encoder.fit_transform([test_y])

    return train_x, train_y, test_x, test_y


def get_class_accuracy(predictor, x, y_expect) -> tuple[List[float], List[float]]:
    from collections import Counter
    y_pred = argmax(predictor.predict(x), axis=1)
    true_positive = [0] * len(set(y_expect))
    class_counter = Counter(list(y_expect))
    class_frequency: List[float] = [class_counter[name] for name, _ in sorted(class_counter.items(), key=lambda i: i[1], reverse=True)]
    for i in range(len(y_expect)):
        pred_class = y_pred[i]
        if pred_class == y_expect[i]:
            true_positive[pred_class] += 1
    return [true_positive[i] / class_frequency[i] for i in range(len(class_frequency))],\
           [i/sum(class_frequency) for i in class_frequency]


def get_f1(predictor, x, y_expect) -> float:
    y_pred = argmax(predictor.predict(x), axis=1)
    return f1_score(y_expect, y_pred, average='macro')



from __future__ import annotations
from typing import Iterable, Callable, List, Any
from tensorflow import Tensor, stack
from tensorflow.keras import Model
from tensorflow.keras.layers import Concatenate, Lambda
from tensorflow.keras.layers import Dense
from tensorflow.keras.models import clone_model
from tensorflow.python.keras.saving.save import load_model
from psyki.logic.datalog import Lukasiewicz, SubNetworkBuilder
from psyki.ski import Injector, Formula
from psyki.utils import eta, eta_one_abs, eta_abs_one


class LambdaLayer(Injector):

    def __init__(self, predictor: Model, class_mapping: dict[str, int],
                 feature_mapping: dict[str, int], gamma: float = 1.):
        self.predictor: Model = predictor
        self.class_mapping: dict[str, int] = class_mapping
        self.feature_mapping: dict[str, int] = feature_mapping
        self.gamma: float = gamma
        self._fuzzy_functions: Iterable[Callable] = ()

    def inject(self, rules: List[Formula]) -> Model:
        fuzzifier = Lukasiewicz(self.class_mapping, self.feature_mapping)
        dict_functions = fuzzifier.visit(rules)
        self._fuzzy_functions = [dict_functions[name] for name, _ in sorted(self.class_mapping.items(),
                                                                            key=lambda i: i[1])]
        predictor_output = self.predictor.layers[-1].output
        x = Concatenate(axis=1)([self.predictor.input, predictor_output])
        x = Lambda(self._cost, self.predictor.output.shape)(x)
        self.predictor = Model(self.predictor.input, x)
        return self.predictor

    def _cost(self, output_layer: Tensor) -> Tensor:
        input_len = self.predictor.input.shape[1]
        x, y = output_layer[:, :input_len], output_layer[:, input_len:]
        cost = stack([function(x, y) for function in self._fuzzy_functions], axis=1)
        return y + (cost / self.gamma)

    def remove(self) -> Any:
        """
        Remove the constraining obtained by the injected rules.
        """
        self.predictor = Model(self.predictor.input, self.predictor.layers[-3].output)
        return self.predictor

    def load(self, file):
        return load_model(file, custom_objects={'_cost': self._cost})


class NetworkComposer(Injector):

    def __init__(self, predictor: Model, feature_mapping: dict[str, int]):
        self.predictor: Model = predictor
        self.feature_mapping: dict[str, int] = feature_mapping
        self._fuzzy_functions: Iterable[Callable] = ()

    def inject(self, rules: List[Formula]) -> Model:
        predictor_input: Tensor = self.predictor.input
        predictor_output: Tensor = self.predictor.layers[-2].output
        activation: Callable = self.predictor.layers[-1].activation
        output_neurons: int = self.predictor.layers[-1].output.shape[1]
        fuzzifier = SubNetworkBuilder(predictor_input, self.feature_mapping)
        modules = fuzzifier.visit(rules)
        new_predictor = Dense(output_neurons, activation=activation)(Concatenate(axis=1)([predictor_output] + modules))
        self.predictor = Model(predictor_input, new_predictor)
        return self.predictor

    @staticmethod
    def load(file: str):
        return load_model(file, custom_objects={'eta': eta, 'eta_one_abs': eta_one_abs, 'eta_abs_one': eta_abs_one})



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, List


class Injector(ABC):
    """
    An injector is a class that allows a sub-symbolic predictor to exploit prior symbolic knowledge.
    The knowledge is provided via rules in some sort of logic form (e.g. FOL, Skolem, Horn).
    """
    predictor: Any  # Any class that has methods fit and predict

    @abstractmethod
    def inject(self, rules: List[Formula]) -> Any:
        pass


class Fuzzifier(ABC):
    """
    A fuzzifier visits a Formula representing symbolic knowledge to build an injectable fuzzy knowledge object.
    """

    @abstractmethod
    def visit(self, rules: List[Formula]) -> Any:
        pass


class Formula(ABC):
    """
    Visitable data structure that represents symbolic knowledge formula.
    """
    pass



from __future__ import annotations
from abc import ABC
from psyki.ski import Formula


class DatalogFormula(Formula):

    def __init__(self, lhs: DefinitionClause, rhs: Clause, op: str):
        self.lhs: DefinitionClause = lhs
        self.rhs: Clause = rhs
        self.op: str = op


class DefinitionClause(Formula):

    def __init__(self, predication: str, arg: Argument):
        self.predication: str = predication
        self.arg: Argument = arg


class Clause(Formula, ABC):
    pass


class Expression(Clause):

    def __init__(self, lhs: Clause, rhs: Clause, op: str):
        self.lhs: Clause = lhs
        self.rhs: Clause = rhs
        self.op: str = op


class Literal(Clause, ABC):
    pass


class Negation(Literal):

    def __init__(self, predicate: Clause):
        self.predicate: Clause = predicate


class Predicate(Clause, ABC):
    pass


class Unary(Predicate):

    def __init__(self, name: str):
        self.name: str = name


class Nary(Predicate):

    def __init__(self, name: str, arg: Argument):
        self.name: str = name
        self.arg: Argument = arg


class Term(Predicate, ABC):
    pass


class Constant(Term, ABC):
    pass


class Predication(Constant):

    def __init__(self, name: str):
        self.name: str = name


class Number(Constant):

    def __init__(self, value: str):
        self.value: float = float(value)


class Variable(Term):

    def __init__(self, name: str):
        self.name: str = name


class Argument(Formula):

    def __init__(self, term: Term, arg: Argument = None):
        self.term: Term = term
        self.arg: Argument = arg


from setuptools import setup, find_packages
import pathlib
import subprocess
import distutils.cmd

# current directory

here = pathlib.Path(__file__).parent.resolve()

version_file = here / 'VERSION'

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')


def format_git_describe_version(version):
    if '-' in version:
        splitted = version.split('-')
        tag = splitted[0]
        index = f"dev{splitted[1]}"
        return f"{tag}.{index}"
    else:
        return version


def get_version_from_git():
    try:
        process = subprocess.run(["git", "describe"], cwd=str(here), check=True, capture_output=True)
        version = process.stdout.decode('utf-8').strip()
        version = format_git_describe_version(version)
        with version_file.open('w') as f:
            f.write(version)
        return version
    except subprocess.CalledProcessError:
        if version_file.exists():
            return version_file.read_text().strip()
        else:
            return '0.1.0'


version = get_version_from_git()


print(f"Detected version {version} from git describe")


class GetVersionCommand(distutils.cmd.Command):
    """A custom command to get the current project version inferred from git describe."""

    description = 'gets the project version from git describe'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version)


class GenerateAntlr4Parser(distutils.cmd.Command):
    """A custom command to generate an Antlr4 parser class for a given grammar."""

    description = 'generate the Antlr4 parser for a given grammar'
    user_options = [('file=', 'f', 'grammar file name')]

    def initialize_options(self):
        from psyki.resources import PATH
        self.file = str(PATH / 'Datalog.g4')

    def finalize_options(self):
        pass

    def run(self):
        import re
        from os import system, popen
        antlr4_version = re.split(r'=', popen('cat requirements.txt | grep antlr4').read())[1][:-1]
        system('wget https://www.antlr.org/download/antlr-' + antlr4_version + '-complete.jar')
        system('export CLASSPATH="./antlr-' + antlr4_version + '-complete.jar:$CLASSPATH"')
        # system('java -jar ./antlr-4.9.2-complete.jar -Dlanguage=Python3 psyki/resources/Datalog.g4 -visitor -o psyki/resources/dist')
        system('java -jar ./antlr-' + antlr4_version + '-complete.jar -Dlanguage=Python3 ' + self.file + ' -visitor -o psyki/resources/dist')
        system('rm ./antlr-' + antlr4_version + '-complete.jar')


setup(
    name='psyki',  # Required
    version=version,
    description='Python-based implementation of PSyKI, i.e. a Platform for Symbolic Knowledge Injection',
    license='Apache 2.0 License',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/psykei/psyki-python',
    author='Matteo Magnini',
    author_email='matteo.magnini@unibo.it',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Prolog'
    ],
    keywords='symbolic knowledge injection, ski, symbolic ai',  # Optional
    # package_dir={'': 'src'},  # Optional
    packages=find_packages(),  # Required
    include_package_data=True,
    python_requires='>=3.9.0, <3.10',
    install_requires=[
        'antlr4-python3-runtime~=4.9.3',
        'tensorflow~=2.6.2',
        'numpy~=1.19.2',
        'scikit-learn~=1.0.1',
        'pandas~=1.3.4',
        # 'skl2onnx~=1.10.0',
        # 'onnxruntime~=1.9.0'
    ],  # Optional
    zip_safe = False,
    platforms = "Independant",
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/psykei/psyki-python/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/psykei/psyki-python',
    },
    cmdclass={
        'get_project_version': GetVersionCommand,
        'generate_antlr4_parser': GenerateAntlr4Parser,
    },
)